from collections import defaultdict
import re
import numpy as np
import numpy.linalg as lin

# The dataset used here can be downloaded at
# http://ir.dcs.gla.ac.uk/resources/test_collections/time/


class SVD:
    """
    A simple container for the three matrices necessary for the
    latent semantic indexing
    """

    def __init__(self, u, s, vt):
        self.u = u
        self.s = s
        self.vt = vt


def import_dataset():
    """
    This function import all the articles in the TIME corpus,
    returning list of lists where each sub-list contains all the
    terms present in the document as a string.
    """
    articles = []
    with open('TIME.ALL', 'r') as f:
        tmp = []
        for row in f:
            if row.startswith("*TEXT"):
                if tmp != []:
                    articles.append(tmp)
                tmp = []
            else:
                row = re.sub(r'[^a-zA-Z\s]+', '', row)
                tmp += row.split()
    return articles


def make_inverted_index(articles):
    """
    This function builds an inverted index as an hash table (dictionary)
    where the keys are the terms and the values are ordered lists of
    docIDs containing the term.
    """
    index = defaultdict(set)
    for docid, article in enumerate(articles):
        for term in article:
            index[term].add(docid)
    return index


def term_document_matrix(articles):
    """
    Given a set of articles we generate a term-document matrix (containing only
    Boolean values) using a numpy array. Notice that for large document
    collections this matrix can be extremely large.
    We also return the list of terms used, which can be used to answer
    new queries.
    """
    index = make_inverted_index(articles)
    terms = list(index.keys())
    mat = np.zeros((len(terms), len(articles)))
    for i, t in enumerate(terms):
        posting_list = index[t]
        for j in posting_list:
            mat[i][j] = 1
    return mat, terms


def latent_semantic_indexing(mat, n_concepts=30):
    u, s, vt = lin.svd(mat)
    # We now build the "reduced" versions of the matrices
    ur = u[:, 0:n_concepts]  # pseudoterms
    sr = s[0:n_concepts]
    vtr = vt[0:n_concepts, :]  # embedding of the documents
    return SVD(ur, sr, vtr)


def cosine_similarity(v1, v2):
    """
    Computes the cosine similarity between two vectors
    """
    len_v1 = lin.norm(v1)
    len_v2 = lin.norm(v2)
    return np.dot(v1, v2)/(len_v1*len_v2)


def answer_query_helper(svd_mat, terms, query):
    """
    An helper function to compute the ranking of all documents.
    """
    # Computation of the inverse of Sigma
    inverse_sr = lin.inv(np.diag(svd_mat.s))
    # Building a query vector from the query
    query_vector = np.zeros((len(terms),))
    for i, t in enumerate(terms):
        if t in query:
            query_vector[i] = 1
    # Compute the embedding of the query vector by multiplying
    # it by \Sigma^-1 * U^T
    embedded_query = inverse_sr @ svd_mat.u.transpose() @ query_vector
    n_documents = svd_mat.vt.shape[1]
    # Now we have a list of pairs (docID, latent representation)
    documents = [(i, svd_mat.vt[:, i]) for i in range(0, n_documents)]

    def scoring_f(x):
        # Scoring is given by computing the cosine similarity
        return (x[0], cosine_similarity(x[1], embedded_query))

    # Computing the scores for all documents
    scores = map(scoring_f, documents)
    # ans sortign them according to cosine similarity
    ranked = sorted(scores, key=lambda x: x[1], reverse=True)
    return ranked


def answer_query(svd_mat, terms, articles, query):
    ranking = answer_query_helper(svd_mat, terms, query)
    for i in range(0, 10):
        article = articles[ranking[i][0]]
        if (len(article) > 30):
            article = article[0:30]
        text = " ".join(article)
        print(f"Article {i + 1}, score: {ranking[i][1]}")
        print(text)


def visualize_pseudoterm(svd_mat, term, idx):
    weights = svd_mat.u[:, idx]
    pt = list(zip(term, weights))
    return pt

# Example of usage
# articles = import_dataset()
# mat, terms = term_document_matrix(articles)
# svd_mat = latent_semantic_indexing(mat, n_concepts=40)
# answer_query(svd_mat, terms, articles, ["FRANCE"])
